from .analysis import Analyzer, Analysis, analyze_json_str, analyze_xml_str
from .analysis_description import AnalysisDescription
from .body_description import BodyDescription
from .case import Case, CaseListener, CaseResult
from .extraction import (
    Extractor,
    JqExtractor,
    XPathExtractor,
    ExtractionError,
)
from .matcher import Matcher, StaticMatcher, ValueMatcher, RecursiveMatcher
from .predicate import Predicate, MatcherPredicate
from .request import Request, RequestParameters, RequestParameterValue
from .response_description import ResponseDescription, ResponseVerification
from .scenario import Scenario, ScenarioListener, ScenarioResult, ScenarioTask
from .status import Status, StatusedMixin, StatusedList
from .type import ScalarType, is_scalar
from .verification import Verification

__all__ = [
    'Analyzer',
    'Analysis',
    'analyze_json_str',
    'analyze_xml_str',
    'AnalysisDescription',
    'BodyDescription',
    'Case',
    'CaseListener',
    'CaseResult',
    'Extractor',
    'JqExtractor',
    'XPathExtractor',
    'ExtractionError',
    'Matcher',
    'StaticMatcher',
    'ValueMatcher',
    'RecursiveMatcher',
    'Predicate',
    'MatcherPredicate',
    'Request',
    'RequestParameters',
    'RequestParameterValue',
    'ResponseDescription',
    'ResponseVerification',
    'Scenario',
    'ScenarioTask',
    'ScenarioListener',
    'ScenarioResult',
    'Status',
    'StatusedMixin',
    'StatusedList',
    'ScalarType',
    'is_scalar',
    'Verification',
]



from __future__ import annotations

from typing import List

from preacher.core.response import Response
from preacher.core.scenario import ScenarioResult
from preacher.presentation.report import Reporter
from . import Listener


class ReportingListener(Listener):

    def __init__(self, reporter: Reporter):
        self._reporter = reporter
        self._results: List[ScenarioResult] = []

    def on_response(self, response: Response) -> None:
        self._reporter.export_response(response)

    def on_scenario(self, result: ScenarioResult) -> None:
        self._results.append(result)

    def on_end(self) -> None:
        self._reporter.export_results(self._results)

    @staticmethod
    def from_path(path: str) -> ReportingListener:
        reporter = Reporter(path)
        return ReportingListener(reporter)



"""
Descriptions, which extract a value from an analyzer and test
along the given predicates.
"""

from typing import Any, List

from .analysis import Analyzer
from .extraction import Extractor
from .predicate import Predicate
from .verification import Verification, collect


class AnalysisDescription:

    def __init__(self, extractor: Extractor, predicates: List[Predicate]):
        self._extractor = extractor
        self._predicates = predicates

    def verify(self, analyzer: Analyzer, **kwargs: Any) -> Verification:
        """`**kwargs` will be delegated to predicates."""
        try:
            verified_value = self._extractor.extract(analyzer)
        except Exception as error:
            return Verification.of_error(error)

        return collect(
            predicate.verify(verified_value, **kwargs)
            for predicate in self._predicates
        )

    @property
    def extractor(self) -> Extractor:
        return self._extractor

    @property
    def predicates(self) -> List[Predicate]:
        return self._predicates



"""
Predicates, which tests a given value.
"""

from abc import ABC, abstractmethod

from .matcher import Matcher, match
from .verification import Verification


class Predicate(ABC):
    """Predicate interface."""

    @abstractmethod
    def verify(self, actual: object, **kwargs) -> Verification:
        raise NotImplementedError()


class MatcherPredicate(Predicate):
    """Predicate of a Hamcrest matcher."""

    def __init__(self, matcher: Matcher):
        self._matcher = matcher

    def verify(self, actual: object, **kwargs) -> Verification:
        return match(self._matcher, actual, **kwargs)



from concurrent.futures import ThreadPoolExecutor
from typing import Iterable, Optional

from .listener import Listener
from .scenario import Scenario
from .scenario.status import Status


class ScenarioRunner:

    def __init__(
        self,
        base_url: str = '',
        retry: int = 0,
        delay: float = 0.1,
        timeout: Optional[float] = None,
    ):
        self._base_url = base_url
        self._retry = retry
        self._delay = delay
        self._timeout = timeout

    def run(
        self,
        executor: ThreadPoolExecutor,
        scenarios: Iterable[Scenario],
        listener: Optional[Listener] = None,
    ) -> Status:
        listener = listener or Listener()
        tasks = [
            scenario.submit(
                executor,
                base_url=self._base_url,
                retry=self._retry,
                delay=self._delay,
                timeout=self._timeout,
                listener=listener,
            )
            for scenario in scenarios
        ]
        results = (task.result() for task in tasks)

        status = Status.SKIPPED
        for result in results:
            status = status.merge(result.status)
            listener.on_scenario(result)

        listener.on_end()
        return status



from .analysis import AnalysisCompiler
from .body_description import BodyDescriptionCompiler
from .case import CaseCompiler
from .analysis_description import AnalysisDescriptionCompiler
from .extraction import ExtractionCompiler
from .predicate import PredicateCompiler
from .request import RequestCompiler
from .response_description import ResponseDescriptionCompiler
from .scenario import ScenarioCompiler


def create_compiler() -> ScenarioCompiler:
    request = RequestCompiler()

    extraction = ExtractionCompiler()
    analysis = AnalysisCompiler()
    predicate = PredicateCompiler()
    description = AnalysisDescriptionCompiler(
        extraction_compiler=extraction,
        predicate_compiler=predicate,
    )
    body_description = BodyDescriptionCompiler(
        analysis=analysis,
        description=description,
    )
    response = ResponseDescriptionCompiler(
        predicate=predicate,
        description=description,
        body=body_description,
    )
    case = CaseCompiler(
        request=request,
        response=response,
    )
    scenario_compiler = ScenarioCompiler(
        description=description,
        case=case,
    )

    return scenario_compiler



from __future__ import annotations

import logging

from preacher.core.scenario import ScenarioResult
from preacher.presentation.log import Logger
from . import Listener


class LoggingListener(Listener):

    def __init__(self, reporter: Logger):
        self._reporter = reporter

    def on_scenario(self, result: ScenarioResult) -> None:
        self._reporter.show_scenario_result(result)

    @staticmethod
    def from_logger(logger: logging.Logger) -> LoggingListener:
        reporter = Logger(logger)
        return LoggingListener(reporter)



from __future__ import annotations

from preacher.core.scenario import ScenarioListener, ScenarioResult


class Listener(ScenarioListener):
    """
    Listener interface.
    Default implementations do nothing.
    """

    def on_end(self) -> None:
        pass

    def on_scenario(self, result: ScenarioResult) -> None:
        pass



"""Request."""

import uuid
from copy import copy
from datetime import datetime
from typing import List, Mapping, Optional, Union

import requests

from preacher import __version__ as _version
from preacher.core.datetime import now
from preacher.core.response import Response, ResponseBody
from .type import ScalarType

_DEFAULT_HEADERS = {'User-Agent': f'Preacher {_version}'}

RequestParameterValue = Union[None, ScalarType, List[Optional[ScalarType]]]
RequestParameters = Union[
    None,
    ScalarType,
    Mapping[str, RequestParameterValue],
]


class ResponseBodyWrapper(ResponseBody):

    def __init__(self, res: requests.Response):
        self._res = res

    @property
    def text(self) -> str:
        return self._res.text

    @property
    def content(self) -> bytes:
        return self._res.content


class ResponseWrapper(Response):

    def __init__(self, id: str, starts: datetime, res: requests.Response):
        self._id = id
        self._starts = starts
        self._res = res
        self._body = ResponseBodyWrapper(self._res)

    @property
    def id(self) -> str:
        return self._id

    @property
    def starts(self) -> datetime:
        return self._starts

    @property
    def elapsed(self) -> float:
        return self._res.elapsed.total_seconds()

    @property
    def status_code(self) -> int:
        return self._res.status_code

    @property
    def headers(self) -> Mapping[str, str]:
        # Convert to the normal dictionary to adapt jq.
        # Names are converted to lower case to normalize.
        return {
            name.lower(): value for (name, value) in self._res.headers.items()
        }

    @property
    def body(self) -> ResponseBody:
        return self._body


class Request:

    def __init__(
        self,
        path: str = '',
        headers: Optional[Mapping[str, str]] = None,
        params: RequestParameters = None,
    ):
        self._path = path
        self._headers = headers or {}
        self._params = params or {}

    def __call__(
        self,
        base_url: str,
        timeout: Optional[float] = None,
    ) -> Response:
        headers = copy(_DEFAULT_HEADERS)
        headers.update(self._headers)
        starts = now()

        res = requests.get(
            base_url + self._path,
            headers=headers,
            params=self._params,  # type: ignore
            timeout=timeout,
        )
        return ResponseWrapper(id=str(uuid.uuid4()), starts=starts, res=res)

    @property
    def path(self) -> str:
        return self._path

    @property
    def headers(self) -> Mapping:
        return self._headers

    @property
    def params(self) -> RequestParameters:
        return self._params



"""Extraction."""

from abc import ABC, abstractmethod
from typing import Any, Callable, List, TypeVar

import pyjq as jq
from lxml.etree import _Element as Element, XPathEvalError

from preacher.core.functional import identify
from .analysis import Analyzer

T = TypeVar('T')


class ExtractionError(RuntimeError):
    pass


class Extractor(ABC):

    @abstractmethod
    def extract(self, analyzer: Analyzer) -> object:
        """
        Raises:
            EvaluationError: when the evaluation of this extractor fails.
        """
        raise NotImplementedError()


class JqExtractor(Extractor):

    def __init__(
        self,
        query: str,
        multiple: bool = False,
        cast: Callable[[object], Any] = identify,
    ):
        self._query = query
        self._multiple = multiple
        self._cast = cast

    def extract(self, analyzer: Analyzer) -> object:
        try:
            compiled = jq.compile(self._query)
        except ValueError:
            raise ExtractionError(f'Invalid jq script: {self._query}')

        values = (
            self._cast(value) if value is not None else value
            for value in analyzer.jq(compiled.all)
        )
        if self._multiple:
            return list(values)
        else:
            return next(values, None)


class XPathExtractor(Extractor):

    def __init__(
        self,
        query: str,
        multiple: bool = False,
        cast: Callable[[object], Any] = identify,
    ):
        self._query = query
        self._multiple = multiple
        self._cast = cast

    def extract(self, analyzer: Analyzer) -> object:
        elements = analyzer.xpath(self._extract)
        if not elements:
            return None

        values = (self._cast(self._convert(element)) for element in elements)
        if self._multiple:
            return list(values)
        else:
            return next(values, None)

    @staticmethod
    def _convert(elem: object) -> str:
        if isinstance(elem, Element):
            return elem.text
        return str(elem)

    def _extract(self, elem: Element) -> List[Element]:
        try:
            return elem.xpath(self._query)
        except XPathEvalError:
            raise ExtractionError(f'Invalid XPath: {self._query}')



"""Request compilation."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Optional

from preacher.core.scenario import (
    Request,
    RequestParameters,
    RequestParameterValue,
    ScalarType,
    is_scalar,
)
from .error import CompilationError, on_key
from .util import compile_str, compile_mapping, map_compile, or_else

_KEY_PATH = 'path'
_KEY_HEADERS = 'headers'
_KEY_PARAMS = 'params'


@dataclass(frozen=True)
class RequestCompiled:
    path: Optional[str] = None
    headers: Optional[Mapping] = None
    params: Optional[RequestParameters] = None

    def replace(self, other: RequestCompiled) -> RequestCompiled:
        return RequestCompiled(
            path=or_else(other.path, self.path),
            headers=or_else(other.headers, self.headers),
            params=or_else(other.params, self.params),  # type: ignore
        )

    def fix(self) -> Request:
        return Request(
            path=or_else(self.path, ''),
            headers=self.headers,
            params=self.params,
        )


class RequestCompiler:

    def __init__(self, default: RequestCompiled = None):
        self._default = default or RequestCompiled()

    def compile(self, obj) -> RequestCompiled:
        """`obj` should be a mapping or a string."""

        if isinstance(obj, str):
            return self.compile({_KEY_PATH: obj})

        obj = compile_mapping(obj)
        compiled = self._default

        path_obj = obj.get(_KEY_PATH)
        if path_obj is not None:
            with on_key(_KEY_PATH):
                path = compile_str(path_obj)
            compiled = replace(compiled, path=path)

        headers_obj = obj.get(_KEY_HEADERS)
        if headers_obj is not None:
            with on_key(_KEY_HEADERS):
                headers = compile_mapping(headers_obj)
            compiled = replace(compiled, headers=headers)

        params_obj = obj.get(_KEY_PARAMS)
        if params_obj is not None:
            with on_key(_KEY_PARAMS):
                params = _compile_params(params_obj)
            compiled = replace(compiled, params=params)

        return compiled

    def of_default(self, default: RequestCompiled) -> RequestCompiler:
        return RequestCompiler(default=self._default.replace(default))


def _compile_param_value_item(item: object) -> Optional[ScalarType]:
    if item is None:
        return item
    if is_scalar(item):
        return item  # type: ignore
    raise CompilationError('Must be a scalar')


def _compile_param_value(value: object) -> RequestParameterValue:
    if value is None:
        return value
    if is_scalar(value):
        return value  # type: ignore

    if not isinstance(value, list):
        raise CompilationError('Must be a scalar or a list')
    return list(map_compile(_compile_param_value_item, value))


def _compile_params(params: object) -> RequestParameters:
    if isinstance(params, str):
        return params

    if not isinstance(params, Mapping):
        raise CompilationError('Must be a string or a map')
    compiled = {}
    for key, value in params.items():
        if not isinstance(key, str):
            raise CompilationError(
                f'A parameter key must be a string, given {key}'
            )
        with on_key(key):
            compiled[key] = _compile_param_value(value)
    return compiled



"""Predicate compilation."""

from preacher.core.scenario import Predicate, MatcherPredicate
from .matcher import compile as compile_matcher


class PredicateCompiler:

    def compile(self, obj: object) -> Predicate:
        matcher = compile_matcher(obj)
        return MatcherPredicate(matcher)



"""
Timedelta interpretation.
"""

import re
from datetime import timedelta

from .error import InterpretationError

TIMEDELTA_PATTERN = re.compile(r'([+\-]?\d+)\s*(day|hour|minute|second)s?')


def interpret_timedelta(value: object) -> timedelta:
    """
    Args:
        value: The interpreted value, which should be a string.

    Raises:
        InterpretationError: When interpretation fails.
    """
    if not isinstance(value, str):
        raise InterpretationError(f'Must be a string, given {type(value)}')

    normalized = value.strip().lower()

    if normalized == 'now':
        return timedelta()

    match = TIMEDELTA_PATTERN.match(normalized)
    if not match:
        raise InterpretationError(f'Invalid timedelta format: {value}')
    offset = int(match.group(1))
    unit = match.group(2) + 's'
    return timedelta(**{unit: offset})



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import List, Optional

from preacher.core.scenario import (
    Analysis,
    analyze_json_str,
    AnalysisDescription,
    BodyDescription,
)
from .analysis import AnalysisCompiler
from .analysis_description import AnalysisDescriptionCompiler
from .error import CompilationError, on_key
from .util import map_compile, or_else

_KEY_ANALYSIS = 'analyze_as'
_KEY_DESCRIPTIONS = 'descriptions'


@dataclass(frozen=True)
class BodyDescriptionCompiled:
    analyze: Optional[Analysis] = None
    descriptions: Optional[List[AnalysisDescription]] = None

    def replace(
        self,
        other: BodyDescriptionCompiled,
    ) -> BodyDescriptionCompiled:
        return BodyDescriptionCompiled(
            analyze=or_else(other.analyze, self.analyze),
            descriptions=or_else(other.descriptions, self.descriptions),
        )

    def fix(self) -> BodyDescription:
        return BodyDescription(
            analyze=self.analyze or analyze_json_str,
            descriptions=self.descriptions,
        )


class BodyDescriptionCompiler:

    def __init__(
        self,
        analysis: AnalysisCompiler,
        description: AnalysisDescriptionCompiler,
        default: Optional[BodyDescriptionCompiled] = None,
    ):
        self._analysis = analysis
        self._description = description
        self._default = default or BodyDescriptionCompiled()

    def compile(self, obj: object) -> BodyDescriptionCompiled:
        """
        `obj` should be a mapping or a list.
        An empty list results in an empty description.
        """

        if isinstance(obj, list):
            return self.compile({_KEY_DESCRIPTIONS: obj})
        if not isinstance(obj, Mapping):
            message = f'Must be a map or a list, given {type(obj)}'
            raise CompilationError(message)

        compiled = self._default

        analyze_obj = obj.get(_KEY_ANALYSIS)
        if analyze_obj is not None:
            with on_key(_KEY_ANALYSIS):
                analyze = self._analysis.compile(analyze_obj)
            compiled = replace(compiled, analyze=analyze)

        descriptions_obj = obj.get(_KEY_DESCRIPTIONS)
        if descriptions_obj is not None:
            descriptions = self._compile_descriptions(descriptions_obj)
            compiled = replace(compiled, descriptions=descriptions)

        return compiled

    def of_default(
        self,
        default: BodyDescriptionCompiled
    ) -> BodyDescriptionCompiler:
        return BodyDescriptionCompiler(
            analysis=self._analysis,
            description=self._description,
            default=self._default.replace(default),
        )

    def _compile_descriptions(self, obj: object) -> List[AnalysisDescription]:
        if not isinstance(obj, list):
            obj = [obj]

        with on_key(_KEY_DESCRIPTIONS):
            return list(map_compile(self._description.compile, obj))



"""
Response body analysis.
"""

import json
from abc import ABC, abstractmethod
from typing import Any, Callable, TypeVar

from lxml.etree import _Element as Element, XMLParser, fromstring

from preacher.core.response import ResponseBody

T = TypeVar('T')


class Analyzer(ABC):
    """
    Interface to analyze body.
    """

    @abstractmethod
    def jq(self, extract: Callable[[object], T]) -> T:
        raise NotImplementedError()

    @abstractmethod
    def xpath(self, extract: Callable[[Element], T]) -> T:
        raise NotImplementedError()


class JsonAnalyzer(Analyzer):

    def __init__(self, json_body: Any):
        self._json_body = json_body

    def jq(self, extract: Callable[[Any], T]) -> T:
        return extract(self._json_body)

    def xpath(self, extract: Callable[[Element], T]) -> T:
        raise NotImplementedError('XPath extraction is not allowed for JSON')


class XmlAnalyzer(Analyzer):

    def __init__(self, etree: Element):
        self._etree = etree

    def jq(self, extract: Callable[[Any], T]) -> T:
        raise NotImplementedError('jq extraction is not allowed for XML')

    def xpath(self, extract: Callable[[Element], T]) -> T:
        return extract(self._etree)


def analyze_json_str(value: ResponseBody) -> Analyzer:
    return JsonAnalyzer(json.loads(value.text))


def analyze_xml_str(value: ResponseBody) -> Analyzer:
    etree = fromstring(value.content, parser=XMLParser())
    return XmlAnalyzer(etree)


Analysis = Callable[[ResponseBody], Analyzer]



"""Extraction compilation."""
from collections.abc import Mapping
from typing import Any, Callable

from preacher.core.functional import identify
from preacher.core.scenario import Extractor, JqExtractor, XPathExtractor
from .error import CompilationError, on_key
from .util import compile_bool, compile_str

_EXTRACTION_MAP = {
    'jq': JqExtractor,
    'xpath': XPathExtractor,
}
_CAST_FUNC_MAP = {
    'int': int,
    'float': float,
    'string': str,
}
_EXTRACTION_KEYS = frozenset(_EXTRACTION_MAP.keys())
_KEY_MULTIPLE = 'multiple'
_KEY_CAST_TO = 'cast_to'


class ExtractionCompiler:

    def compile(self, obj: object) -> Extractor:
        """`obj` should be a mapping or a string."""

        if isinstance(obj, str):
            return self.compile({'jq': obj})

        if not isinstance(obj, Mapping):
            message = f'Must be a map or a string, given {type(obj)}'
            raise CompilationError(message)

        keys = _EXTRACTION_KEYS.intersection(obj.keys())
        if len(keys) != 1:
            raise CompilationError(
                f'Must have only 1 valid key, but has {len(keys)}'
            )
        key = next(iter(keys))

        func = _EXTRACTION_MAP[key]
        query = obj[key]

        multiple_obj = obj.get(_KEY_MULTIPLE, False)
        with on_key(_KEY_MULTIPLE):
            multiple = compile_bool(multiple_obj)

        cast: Callable[[object], Any] = identify
        cast_obj = obj.get(_KEY_CAST_TO)
        if cast_obj is not None:
            with on_key(_KEY_CAST_TO):
                cast = self._compile_cast(cast_obj)

        return func(query, multiple=multiple, cast=cast)

    @staticmethod
    def _compile_cast(obj: object) -> Callable[[object], Any]:
        """`obj` should be a string."""

        key = compile_str(obj)
        cast = _CAST_FUNC_MAP.get(key)
        if not cast:
            raise CompilationError(f'Invalid value: {key}')

        return cast



from preacher.core.scenario import Analysis, analyze_json_str, analyze_xml_str
from .error import CompilationError
from .util import compile_str

_ANALYSIS_MAP = {
    'json': analyze_json_str,
    'xml': analyze_xml_str,
}


class AnalysisCompiler:

    @staticmethod
    def compile(obj: object) -> Analysis:
        """`obj` should be a string."""

        key = compile_str(obj)
        analysis = _ANALYSIS_MAP.get(key)
        if not analysis:
            raise CompilationError(f'Invalid key: {key}')

        return analysis



from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Optional

from preacher.core.datetime import now
from .analysis import Analyzer, JsonAnalyzer


@dataclass(frozen=True)
class ScenarioContext:
    starts: datetime = field(default_factory=now)
    base_url: str = ''
    retry: int = 0
    delay: float = 0.1
    timeout: Optional[float] = None


def _to_serializable(value: object) -> object:
    if isinstance(value, dict):
        return {k: _to_serializable(v) for (k, v) in value.items()}
    if isinstance(value, list):
        return [_to_serializable(v) for v in value]
    if isinstance(value, datetime):
        return value.isoformat()
    return value


def analyze_context(context) -> Analyzer:
    return JsonAnalyzer(_to_serializable(asdict(context)))



"""Matchers."""

from abc import ABC, abstractmethod
from typing import Callable, Generic, List, TypeVar

from hamcrest import assert_that
from hamcrest.core.matcher import Matcher as HamcrestMatcher

from preacher.core.functional import identify
from preacher.core.interpretation.value import Value
from .status import Status
from .verification import Verification

T = TypeVar('T')


class Matcher(ABC):
    """
    Matcher interfaces.
    Matchers are implemented as factories of Hamcrest matchers.
    """

    @abstractmethod
    def to_hamcrest(self, **kwargs) -> HamcrestMatcher:
        raise NotImplementedError()


class StaticMatcher(Matcher):

    def __init__(self, hamcrest: HamcrestMatcher):
        self._hamcrest = hamcrest

    def to_hamcrest(self, **kwargs) -> HamcrestMatcher:
        return self._hamcrest


class ValueMatcher(Matcher, Generic[T]):

    def __init__(
        self,
        hamcrest_factory: Callable[..., HamcrestMatcher],
        value: Value[T],
        interpret: Callable = identify,
    ):
        self._hamcrest_factory = hamcrest_factory
        self._value = value
        self._interpret = interpret

    def to_hamcrest(self, **kwargs) -> HamcrestMatcher:
        value_in_context = self._value.apply_context(**kwargs)
        value_interpreted = self._interpret(value_in_context, **kwargs)
        return self._hamcrest_factory(value_interpreted)


class RecursiveMatcher(Matcher):

    def __init__(
        self,
        hamcrest_factory: Callable[..., HamcrestMatcher],
        inner_matchers: List[Matcher],
    ):
        self._hamcrest_factory = hamcrest_factory
        self._inner_matchers = inner_matchers

    def to_hamcrest(self, **kwargs) -> HamcrestMatcher:
        inner_hamcrest_matchers = (
            inner_matcher.to_hamcrest(**kwargs)
            for inner_matcher in self._inner_matchers
        )
        return self._hamcrest_factory(*inner_hamcrest_matchers)


def match(matcher: Matcher, actual: object, **kwargs) -> Verification:
    try:
        hamcrest_matcher = matcher.to_hamcrest(**kwargs)
        assert_that(actual, hamcrest_matcher)
    except AssertionError as error:
        message = str(error).strip()
        return Verification(status=Status.UNSTABLE, message=message)
    except Exception as error:
        return Verification.of_error(error)

    return Verification.succeed()



"""Utilities for compilations."""

from collections.abc import Mapping
from functools import partial
from typing import Any, Callable, Iterable, Iterator, Optional, TypeVar

from .error import CompilationError, on_key, on_index

T = TypeVar('T')
U = TypeVar('U')


def compile_bool(obj: object) -> bool:
    """
    Compile the given boolean object.

    Args:
        obj: A compiled object, which should be a `bool` value.
    Returns:
        The compiled value.
    Raises:
        CompilationError: when compilation fails.
    """
    if not isinstance(obj, bool):
        raise CompilationError(f'Must be a boolean, given {type(obj)}')
    return obj


def compile_str(obj: object) -> str:
    """
    Compile the given string object.

    Args:
        obj: A compiled object, which should be a `string` value.
    Returns:
        The compiled value.
    Raises:
        CompilationError: when compilation fails.
    """
    if not isinstance(obj, str):
        raise CompilationError(f'must be a string, given {type(obj)}')
    return obj


def compile_optional_str(obj: object) -> Optional[str]:
    """
    Compile the given optional string object.

    Args:
        obj: A compiled object, which should be a `string` value or `None`.
    Returns:
        The compiled value.
    Raises:
        CompilationError: when compilation fails.
    """
    if obj is None:
        return None
    return compile_str(obj)


def compile_list(obj: object) -> list:
    if not isinstance(obj, list):
        return [obj]
    return obj


def compile_mapping(obj: object) -> Mapping:
    """
    Compile the given mapping object.

    Args:
        obj: A compiled object, which should be a mapping.
    Returns:
        The compile result.
    Raises:
        CompilationError: when compilation fails.
    """
    if not isinstance(obj, Mapping):
        raise CompilationError(f'Must be a map, given {type(obj)}')
    return obj


def map_compile(func: Callable[[T], U], items: Iterable[T]) -> Iterator[U]:
    for idx, item in enumerate(items):
        with on_index(idx):
            yield func(item)


def run_recursively(func: Callable[[object], Any], obj) -> object:
    if isinstance(obj, Mapping):
        def _func(key: object, value: object) -> object:
            if not isinstance(key, str):
                raise CompilationError(
                    f'Key must be a string, given {type(key)}'
                )
            with on_key(key):
                return run_recursively(func, value)

        return {k: _func(k, v) for (k, v) in obj.items()}

    if isinstance(obj, list):
        _func = partial(run_recursively, func)
        return list(map_compile(_func, obj))
    return func(obj)


def or_else(optional: Optional[T], default: T) -> T:
    if optional is None:
        return default
    return optional



from __future__ import annotations

import os
from typing import Union, TextIO

from ruamel.yaml import YAML, Node
from ruamel.yaml.error import MarkedYAMLError

from .argument import ArgumentValue
from .error import CompilationError
from .util import run_recursively

PathLike = Union[str, os.PathLike]


class _Inclusion:
    yaml_tag = '!include'

    def __init__(self, obj: object):
        self._obj = obj

    def resolve(self, origin: PathLike, yaml: YAML) -> object:
        obj = self._obj
        if not isinstance(obj, str):
            raise CompilationError(f'Must be a string, given {type(obj)}')

        path = os.path.join(origin, obj)
        with open(path) as f:
            return _load(yaml, f, os.path.dirname(path))

    @classmethod
    def from_yaml(cls, _constructor, node: Node) -> _Inclusion:
        return _Inclusion(node.value)


class _ArgumentValue:
    yaml_tag = '!argument'

    def __init__(self, obj: object):
        self._obj = obj

    def resolve(self) -> ArgumentValue:
        obj = self._obj
        if not isinstance(obj, str):
            raise CompilationError(f'Must be a key string, given {type(obj)}')
        return ArgumentValue(obj)

    @classmethod
    def from_yaml(cls, _constructor, node: Node) -> _ArgumentValue:
        return _ArgumentValue(node.value)


def _resolve(yaml: YAML, obj: object, origin: PathLike) -> object:
    if isinstance(obj, _Inclusion):
        return obj.resolve(origin, yaml)

    if isinstance(obj, _ArgumentValue):
        return obj.resolve()

    return obj


def _load(yaml: YAML, io: TextIO, origin: PathLike) -> object:
    try:
        obj = yaml.load(io)
    except MarkedYAMLError as error:
        raise CompilationError(message=str(error), cause=error)

    return run_recursively(lambda o: _resolve(yaml, o, origin), obj)


def load(io: TextIO, origin: PathLike = '.') -> object:
    yaml = YAML(typ='safe', pure=True)
    yaml.register_class(_Inclusion)
    yaml.register_class(_ArgumentValue)
    return _load(yaml, io, origin)


def load_from_path(path: PathLike) -> object:
    origin = os.path.dirname(path)
    with open(path) as f:
        return load(f, origin)



"""
Hamcrest custom matchers.
"""

from datetime import datetime
from typing import Callable, Optional

import hamcrest
from hamcrest.core.base_matcher import BaseMatcher
from hamcrest.core.matcher import Matcher

from preacher.core.datetime import parse_datetime


class _ConvertingMatcher(BaseMatcher):

    def __init__(self, matcher: BaseMatcher, convert: Callable):
        self._matcher = matcher
        self._convert = convert

    def _matches(self, item):
        converted_item = self._convert(item)
        return self._matcher._matches(converted_item)

    def describe_to(self, description):
        self._matcher.describe_to(description)

    def describe_mismatch(self, item, mismatch_description):
        converted_item = self._convert(item)
        mismatch_description.append_text('was ').append_description_of(
            converted_item
        )


def _convert_to_datetime(value: object) -> Optional[datetime]:
    if not isinstance(value, str):
        raise TypeError(f'Must be a string, but given {value}')
    return parse_datetime(value)


def _string_datetime_matcher(matcher: BaseMatcher) -> Matcher:
    return _ConvertingMatcher(matcher, _convert_to_datetime)


def before(value) -> Matcher:
    # TODO: Fix typing.
    return _string_datetime_matcher(hamcrest.less_than(value))  # type: ignore


def after(value) -> Matcher:
    # TODO: Fix typing.
    return _string_datetime_matcher(
        hamcrest.greater_than(value)  # type: ignore
    )



from __future__ import annotations

from typing import List

from preacher.core.response import Response
from preacher.core.scenario import ScenarioResult
from . import Listener


class MergingListener(Listener):

    def __init__(self):
        self._listeners: List[Listener] = []

    def append(self, listener: Listener):
        self._listeners.append(listener)

    def on_response(self, response: Response) -> None:
        for listener in self._listeners:
            listener.on_response(response)

    def on_scenario(self, result: ScenarioResult) -> None:
        for listener in self._listeners:
            listener.on_scenario(result)

    def on_end(self):
        for listener in self._listeners:
            listener.on_end()



"""
Response body description.
"""

from typing import List, Optional

from preacher.core.response import ResponseBody
from .analysis import Analysis, analyze_json_str
from .analysis_description import AnalysisDescription
from .verification import Verification, collect


class BodyDescription:

    def __init__(
        self,
        analyze: Analysis = analyze_json_str,
        descriptions: Optional[List[AnalysisDescription]] = None,
    ):
        self._analyze = analyze
        self._descriptions = descriptions or []

    def verify(self, body: ResponseBody, **kwargs) -> Verification:
        try:
            analyzer = self._analyze(body)
        except Exception as error:
            return Verification.of_error(error)

        return collect(
            description.verify(analyzer, **kwargs)
            for description in self._descriptions
        )



"""Matcher compilation."""

from collections.abc import Mapping
from typing import Callable, Dict

import hamcrest
from hamcrest.core.matcher import Matcher as HamcrestMatcher

from preacher.core.functional import identify
from preacher.core.hamcrest import after, before
from preacher.core.interpretation.datetime import interpret_datetime
from preacher.core.interpretation.value import value_of
from preacher.core.scenario import (
    Matcher,
    StaticMatcher,
    ValueMatcher,
    RecursiveMatcher,
)
from .error import CompilationError, on_key
from .util import compile_list, map_compile

_STATIC_MATCHER_MAP = {
    # For objects.
    'be_null': StaticMatcher(hamcrest.is_(hamcrest.none())),
    'not_be_null': StaticMatcher(hamcrest.is_(hamcrest.not_none())),

    # For collections.
    'be_empty': StaticMatcher(hamcrest.is_(hamcrest.empty())),

    # Logical.
    'anything': StaticMatcher(hamcrest.is_(hamcrest.anything())),
}

_VALUE_MATCHER_HAMCREST_MAP = {
    # For objects.
    'equal': hamcrest.equal_to,
    'have_length': hamcrest.has_length,

    # For comparables.
    'be_greater_than': hamcrest.greater_than,
    'be_greater_than_or_equal_to': hamcrest.greater_than_or_equal_to,
    'be_less_than': hamcrest.less_than,
    'be_less_than_or_equal_to': hamcrest.less_than_or_equal_to,

    # For strings.
    'contain_string': hamcrest.contains_string,
    'start_with': hamcrest.starts_with,
    'end_with': hamcrest.ends_with,
    'match_regexp': hamcrest.matches_regexp,

    # For datetime.
    'be_before': before,
    'be_after': after,
}

_SINGLE_MATCHER_HAMCREST_MAP: Dict[
    str,
    Callable[[HamcrestMatcher], HamcrestMatcher]
] = {
    'be': hamcrest.is_,
    'not': hamcrest.not_,
    'have_item': hamcrest.has_item,
}

_MULTI_MATCHERS_HAMCREST_MAP = {
    'contain': hamcrest.contains_exactly,
    'contain_exactly': hamcrest.contains_exactly,
    'contain_in_any_order': hamcrest.contains_inanyorder,
    'have_items': hamcrest.has_items,
    'all_of': hamcrest.all_of,
    'any_of': hamcrest.any_of,
}

_INTERPRETER_MAP = {
    'be_before': interpret_datetime,
    'be_after': interpret_datetime,
}


def _compile_taking_single_matcher(key: str, value: object) -> Matcher:
    hamcrest_factory = _SINGLE_MATCHER_HAMCREST_MAP[key]

    if isinstance(value, str) or isinstance(value, Mapping):
        with on_key(key):
            inner = compile(value)
    else:
        inner = ValueMatcher(hamcrest.equal_to, value_of(value))

    return RecursiveMatcher(hamcrest_factory, [inner])


def _compile_taking_multi_matchers(key: str, value: object) -> Matcher:
    hamcrest_factory = _MULTI_MATCHERS_HAMCREST_MAP[key]

    with on_key(key):
        value = compile_list(value)
        inner_matchers = list(map_compile(compile, value))

    # TODO: Fix typing.
    return RecursiveMatcher(hamcrest_factory, inner_matchers)  # type: ignore


def compile(obj: object) -> Matcher:
    if isinstance(obj, str):
        if obj in _STATIC_MATCHER_MAP:
            return _STATIC_MATCHER_MAP[obj]

    if isinstance(obj, Mapping):
        if len(obj) != 1:
            raise CompilationError(
                f'Must have only 1 element, but has {len(obj)}'
            )

        key, value = next(iter(obj.items()))

        if key in _VALUE_MATCHER_HAMCREST_MAP:
            # TODO: Fix typing.
            return ValueMatcher(
                _VALUE_MATCHER_HAMCREST_MAP[key],  # type: ignore
                value_of(value),
                interpret=_INTERPRETER_MAP.get(key, identify),
            )

        if key in _SINGLE_MATCHER_HAMCREST_MAP:
            return _compile_taking_single_matcher(key, value)

        if key in _MULTI_MATCHERS_HAMCREST_MAP:
            return _compile_taking_multi_matchers(key, value)

    return ValueMatcher(hamcrest.equal_to, value_of(obj))



"""
Datetime interpretation.
"""

from datetime import datetime, timezone

from preacher.core.datetime import now
from .timedelta import interpret_timedelta


def interpret_datetime(value: object, **kwargs) -> datetime:
    """
    Args:
        value (object): The interpreted value,
            which should be a datetime object or a string.
        **kwargs: options.
            - origin_datetime (datetime, optional):
              The origin of relative datetime.

    Raises:
        InterpretationError: When interpretation fails.
    """
    if isinstance(value, datetime):
        if not value.tzinfo:
            return value.replace(tzinfo=timezone.utc)
        return value

    delta = interpret_timedelta(value)
    origin = kwargs.get('origin_datetime') or now()
    return origin + delta



"""Response description compilations."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Optional, List

from preacher.core.scenario import (
    AnalysisDescription,
    Predicate,
    ResponseDescription,
)
from .body_description import BodyDescriptionCompiler, BodyDescriptionCompiled
from .analysis_description import AnalysisDescriptionCompiler
from .error import CompilationError, on_key
from .predicate import PredicateCompiler
from .util import compile_mapping, map_compile, or_else

_KEY_STATUS_CODE = 'status_code'
_KEY_HEADERS = 'headers'
_KEY_BODY = 'body'


@dataclass(frozen=True)
class ResponseDescriptionCompiled:
    status_code: Optional[List[Predicate]] = None
    headers: Optional[List[AnalysisDescription]] = None
    body: Optional[BodyDescriptionCompiled] = None

    def replace(
        self,
        other: ResponseDescriptionCompiled,
    ) -> ResponseDescriptionCompiled:
        return ResponseDescriptionCompiled(
            status_code=or_else(other.status_code, self.status_code),
            headers=or_else(other.headers, self.headers),
            body=or_else(other.body, self.body),
        )

    def fix(self) -> ResponseDescription:
        body = None
        if self.body:
            body = self.body.fix()

        return ResponseDescription(
            status_code=self.status_code,
            headers=self.headers,
            body=body,
        )


class ResponseDescriptionCompiler:

    def __init__(
        self,
        predicate: PredicateCompiler,
        description: AnalysisDescriptionCompiler,
        body: BodyDescriptionCompiler,
        default: Optional[ResponseDescriptionCompiled] = None,
    ):
        self._predicate = predicate
        self._description = description
        self._body = body
        self._default = default or ResponseDescriptionCompiled()

    def compile(self, obj: object) -> ResponseDescriptionCompiled:
        """`obj` should be a mapping."""

        obj = compile_mapping(obj)
        compiled = self._default

        status_code_obj = obj.get(_KEY_STATUS_CODE)
        if status_code_obj is not None:
            with on_key(_KEY_STATUS_CODE):
                status_code = self._compile_status_code(status_code_obj)
            compiled = replace(compiled, status_code=status_code)

        headers_obj = obj.get(_KEY_HEADERS)
        if headers_obj is not None:
            with on_key(_KEY_HEADERS):
                headers = self._compile_headers(headers_obj)
            compiled = replace(compiled, headers=headers)

        body_obj = obj.get(_KEY_BODY)
        if body_obj is not None:
            with on_key(_KEY_BODY):
                body = self._body.compile(body_obj)
            compiled = replace(compiled, body=body)

        return compiled

    def of_default(
        self,
        default: ResponseDescriptionCompiled,
    ) -> ResponseDescriptionCompiler:
        body = self._body
        if default.body:
            body = body.of_default(default.body)

        return ResponseDescriptionCompiler(
            predicate=self._predicate,
            description=self._description,
            body=body,
            default=self._default.replace(default),
        )

    def _compile_status_code(self, obj: object) -> List[Predicate]:
        if not isinstance(obj, list):
            obj = [obj]
        return list(map_compile(self._predicate.compile, obj))

    def _compile_headers(self, obj: object) -> List[AnalysisDescription]:
        if isinstance(obj, Mapping):
            obj = [obj]
        if not isinstance(obj, list):
            message = f'Must be a list or a map, given {type(obj)}'
            raise CompilationError(message)

        return list(map_compile(self._description.compile, obj))



"""Compilation errors."""

from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from functools import reduce
from typing import List, Optional, Union, Iterator


@dataclass(frozen=True)
class NamedNode:
    name: str

    def __str__(self) -> str:
        return f'.{self.name}'


@dataclass(frozen=True)
class IndexedNode:
    index: int

    def __str__(self) -> str:
        return f'[{self.index}]'


Node = Union[NamedNode, IndexedNode]
Path = List[Node]


class CompilationError(Exception):
    """Compilation errors."""

    def __init__(
        self,
        message: str,
        path: Optional[Path] = None,
        cause: Optional[Exception] = None,
    ):
        super().__init__(message)
        self._message = message
        self._path = path or []
        self._cause = cause

    @property
    def path(self) -> Path:
        return self._path

    def of_parent(self, parent_path: Path) -> CompilationError:
        return CompilationError(
            message=self._message,
            path=parent_path + self.path,
            cause=self._cause,
        )

    def __str__(self):
        message = super().__str__()
        if not self._path:
            return message

        path = reduce(lambda lhs, rhs: lhs + str(rhs), self._path, '')
        return f'{message}: {path}'


@contextmanager
def on_key(key: str) -> Iterator:
    try:
        yield
    except CompilationError as error:
        raise error.of_parent([NamedNode(key)])


@contextmanager
def on_index(index: int) -> Iterator:
    try:
        yield
    except CompilationError as error:
        raise error.of_parent([IndexedNode(index)])



import os
from typing import Iterable

import jinja2

from preacher.core.response import Response
from preacher.core.scenario import ScenarioResult


class Reporter:

    def __init__(self, path):
        self._path = path
        self._responses_path = os.path.join(self._path, 'responses')

        env = jinja2.Environment(
            loader=jinja2.PackageLoader('preacher', 'resources/report/html'),
            autoescape=jinja2.select_autoescape(['html', 'xml']),
        )
        self._index_template = env.get_template('index.html')
        self._response_view_template = env.get_template('response-view.html')

        self._initialize()

    def _initialize(self) -> None:
        os.makedirs(self._path, exist_ok=True)
        os.makedirs(self._responses_path, exist_ok=True)

    def export_response(self, response: Response) -> None:
        name = f'{response.id}.html'
        path = os.path.join(self._responses_path, name)
        with open(path, 'w') as f:
            self._response_view_template.stream(response=response).dump(f)

    def export_results(self, results: Iterable[ScenarioResult]) -> None:
        html_path = os.path.join(self._path, 'index.html')
        with open(html_path, 'w') as f:
            self._index_template.stream(scenarios=results).dump(f)



"""
Response descriptions, which verify the status code, the headers
and the body.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List, Mapping, Optional

from preacher.core.response import Response
from .analysis import Analyzer, JsonAnalyzer
from .body_description import BodyDescription
from .analysis_description import AnalysisDescription
from .predicate import Predicate
from .status import Status, merge_statuses
from .verification import Verification, collect


@dataclass(frozen=True)
class ResponseVerification:
    response_id: str
    status: Status
    status_code: Verification
    headers: Verification
    body: Verification


class ResponseDescription:

    def __init__(
        self,
        status_code: Optional[List[Predicate]] = None,
        headers: Optional[List[AnalysisDescription]] = None,
        body: Optional[BodyDescription] = None,
        analyze_headers:
            Callable[[Mapping[str, str]], Analyzer] = JsonAnalyzer,
    ):
        self._status_code = status_code or []
        self._headers = headers or []
        self._body = body
        self._analyze_headers = analyze_headers

    def verify(self, response: Response, **kwargs) -> ResponseVerification:
        """`**kwargs` will be delegated to descriptions."""
        status_code = self._verify_status_code(response.status_code, **kwargs)

        try:
            headers = self._verify_headers(response.headers, **kwargs)
        except Exception as error:
            headers = Verification.of_error(error)

        body = Verification.skipped()
        if self._body:
            body = self._body.verify(response.body, **kwargs)

        status = merge_statuses([
            status_code.status,
            headers.status,
            body.status,
        ])
        return ResponseVerification(
            response_id=response.id,
            status=status,
            status_code=status_code,
            headers=headers,
            body=body,
        )

    def _verify_status_code(self, code: int, **kwargs) -> Verification:
        return collect(
            predicate.verify(code, **kwargs)
            for predicate in self._status_code
        )

    def _verify_headers(
        self,
        headers: Mapping[str, str],
        **kwargs
    ) -> Verification:
        analyzer = self._analyze_headers(headers)
        return collect(
            description.verify(analyzer, **kwargs)
            for description in self._headers
        )



"""Scenario"""

from __future__ import annotations

from abc import ABC, abstractmethod
from concurrent.futures import Executor
from dataclasses import dataclass, field
from typing import Callable, List, Optional

from .case import Case, CaseListener, CaseResult
from .context import ScenarioContext, analyze_context
from .analysis_description import AnalysisDescription
from .status import (
    Status,
    StatusedList,
    StatusedMixin,
    collect_statused,
    merge_statuses,
)
from .util.concurrency import CasesTask, OrderedCasesTask, UnorderedCasesTask
from .verification import Verification, collect


class ScenarioListener(CaseListener):
    """
    Interface to listen to scenario running.
    """
    pass


@dataclass(frozen=True)
class ScenarioResult(StatusedMixin):
    label: Optional[str] = None
    message: Optional[str] = None
    conditions: Verification = field(default_factory=Verification)
    cases: StatusedList[CaseResult] = field(default_factory=StatusedList)
    subscenarios: StatusedList[ScenarioResult] = field(
        default_factory=StatusedList,
    )


class ScenarioTask(ABC):

    @abstractmethod
    def result(self) -> ScenarioResult:
        raise NotImplementedError()


class RunningScenarioTask(ScenarioTask):

    def __init__(
        self,
        label: Optional[str],
        conditions: Verification,
        cases: CasesTask,
        subscenarios: List[ScenarioTask],
    ):
        self._label = label
        self._conditions = conditions
        self._cases = cases
        self._subscenarios = subscenarios

    def result(self) -> ScenarioResult:
        cases = self._cases.result()
        subscenarios = collect_statused(s.result() for s in self._subscenarios)
        status = merge_statuses([cases.status, subscenarios.status])
        return ScenarioResult(
            label=self._label,
            status=status,
            conditions=self._conditions,
            cases=cases,
            subscenarios=subscenarios,
        )


class StaticScenarioTask(ScenarioTask):

    def __init__(self, result: ScenarioResult):
        self._result = result

    def result(self) -> ScenarioResult:
        return self._result


class Scenario:

    def __init__(
        self,
        label: Optional[str] = None,
        ordered: bool = True,
        conditions: Optional[List[AnalysisDescription]] = None,
        cases: Optional[List[Case]] = None,
        subscenarios: Optional[List[Scenario]] = None,
    ):
        self._label = label
        self._ordered = ordered
        self._conditions = conditions or []
        self._cases = cases or []
        self._subscenarios = subscenarios or []

    def submit(
        self,
        executor: Executor,
        base_url: str = '',
        retry: int = 0,
        delay: float = 0.1,
        timeout: Optional[float] = None,
        listener: Optional[ScenarioListener] = None,
    ) -> ScenarioTask:
        context = ScenarioContext(
            base_url=base_url,
            retry=retry,
            delay=delay,
            timeout=timeout,
        )
        context_analyzer = analyze_context(context)
        conditions = collect(
            condition.verify(context_analyzer, origin_datetime=context.starts)
            for condition in self._conditions
        )
        if conditions.status == Status.FAILURE:
            return StaticScenarioTask(ScenarioResult(
                label=self._label,
                status=Status.FAILURE,
                conditions=conditions,
            ))
        if conditions.status == Status.UNSTABLE:
            return StaticScenarioTask(ScenarioResult(
                label=self._label,
                status=Status.SKIPPED,
                conditions=conditions,
            ))

        listener = listener or ScenarioListener()

        if self._ordered:
            submit_cases: Callable = OrderedCasesTask
        else:
            submit_cases = UnorderedCasesTask
        cases = submit_cases(
            executor,
            self._cases,
            base_url=base_url,
            retry=retry,
            delay=delay,
            timeout=timeout,
            listener=listener,
        )

        subscenarios = [
            subscenario.submit(
                executor,
                base_url=base_url,
                retry=retry,
                delay=delay,
                timeout=timeout,
                listener=listener,
            )
            for subscenario in self._subscenarios
        ]
        return RunningScenarioTask(
            label=self._label,
            conditions=conditions,
            cases=cases,
            subscenarios=subscenarios,
        )



from abc import ABC, abstractmethod
from concurrent.futures import Executor
from typing import Iterable

from preacher.core.scenario.case import Case, CaseResult
from preacher.core.scenario.status import StatusedList, collect_statused


class CasesTask(ABC):

    @abstractmethod
    def result(self) -> StatusedList[CaseResult]:
        raise NotImplementedError()


def _run_cases_in_order(
    cases: Iterable[Case],
    *args,
    **kwargs
) -> StatusedList[CaseResult]:
    return collect_statused(case.run(*args, **kwargs) for case in cases)


class OrderedCasesTask(CasesTask):

    def __init__(
        self,
        executor: Executor,
        cases: Iterable[Case],
        *args,
        **kwargs,
    ):
        self._future = executor.submit(
            _run_cases_in_order,
            cases,
            *args,
            **kwargs,
        )

    def result(self) -> StatusedList[CaseResult]:
        return self._future.result()


class UnorderedCasesTask(CasesTask):

    def __init__(
        self,
        executor: Executor,
        cases: Iterable[Case],
        *args,
        **kwargs,
    ):
        self._futures = [
            executor.submit(case.run, *args, **kwargs)
            for case in cases
        ]

    def result(self) -> StatusedList[CaseResult]:
        return collect_statused(f.result() for f in self._futures)



"""
Functional utilities.
"""

from typing import TypeVar

T = TypeVar('T')


def identify(value: T, *_args, **_kwargs) -> T:
    """
    Returns the first argument.
    >>> identify(None)
    >>> identify(1)
    1
    >>> identify('a')
    'a'
    >>> identify([1, 2, 3])
    [1, 2, 3]

    Ignores extra arguments.
    >>> identify(1, 2, key='value')
    1
    """
    return value



"""
Datetime utilities for Preacher core.
"""

import datetime
import time

import aniso8601


def now() -> datetime.datetime:
    return datetime.datetime.now(_system_timezone())


def parse_datetime(value: str) -> datetime.datetime:
    try:
        return aniso8601.parse_datetime(value)
    except ValueError:
        raise ValueError(f'An invalid datetime format: {value}')


def _system_timezone() -> datetime.timezone:
    localtime = time.localtime()
    return datetime.timezone(
        offset=datetime.timedelta(seconds=localtime.tm_gmtoff),
        name=localtime.tm_zone,
    )



"""Description compilation."""

from typing import Optional

from preacher.core.scenario import AnalysisDescription
from .error import on_key
from .extraction import ExtractionCompiler
from .predicate import PredicateCompiler
from .util import compile_mapping, map_compile

_KEY_DESCRIBE = 'describe'
_KEY_SHOULD = 'should'


class AnalysisDescriptionCompiler:

    def __init__(
        self,
        extraction_compiler: Optional[ExtractionCompiler] = None,
        predicate_compiler: Optional[PredicateCompiler] = None,
    ):
        self._extraction_compiler = extraction_compiler or ExtractionCompiler()
        self._predicate_compiler = predicate_compiler or PredicateCompiler()

    def compile(self, obj: object):
        """`obj` should be a mapping."""

        obj = compile_mapping(obj)

        extraction_obj = obj.get(_KEY_DESCRIBE)
        with on_key(_KEY_DESCRIBE):
            extractor = self._extraction_compiler.compile(extraction_obj)

        predicate_objs = obj.get(_KEY_SHOULD, [])
        if not isinstance(predicate_objs, list):
            predicate_objs = [predicate_objs]
        with on_key(_KEY_SHOULD):
            predicates = list(map_compile(
                self._predicate_compiler.compile,
                predicate_objs,
            ))

        return AnalysisDescription(extractor=extractor, predicates=predicates)



"""
Value interpretation.
"""

from abc import ABC, abstractmethod
from typing import TypeVar, Generic

T = TypeVar('T')

_KEY_ARGUMENTS = "arguments"


class Value(ABC, Generic[T]):

    @abstractmethod
    def apply_context(self, **kwargs) -> T:
        raise NotImplementedError()


class StaticValue(Value[T]):

    def __init__(self, value: T):
        self._value = value

    def apply_context(self, **kwargs) -> T:
        return self._value


def value_of(obj: T) -> Value[T]:
    return StaticValue(obj)



"""Status."""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass, field
from enum import Enum
from functools import reduce
from typing import Generic, List, TypeVar

T = TypeVar('T')


class Status(Enum):
    """
    >>> Status.SKIPPED.is_succeeded
    True
    >>> bool(Status.SKIPPED)
    True
    >>> Status.SKIPPED.merge(Status.SUCCESS)
    SUCCESS
    >>> Status.SKIPPED.merge(Status.UNSTABLE)
    UNSTABLE
    >>> Status.SKIPPED.merge(Status.FAILURE)
    FAILURE

    >>> Status.SUCCESS.is_succeeded
    True
    >>> bool(Status.SUCCESS)
    True
    >>> Status.SUCCESS.merge(Status.SKIPPED)
    SUCCESS
    >>> Status.SUCCESS.merge(Status.SUCCESS)
    SUCCESS
    >>> Status.SUCCESS.merge(Status.UNSTABLE)
    UNSTABLE
    >>> Status.SUCCESS.merge(Status.FAILURE)
    FAILURE

    >>> Status.UNSTABLE.is_succeeded
    False
    >>> bool(Status.UNSTABLE)
    False
    >>> Status.UNSTABLE.merge(Status.SKIPPED)
    UNSTABLE
    >>> Status.UNSTABLE.merge(Status.SUCCESS)
    UNSTABLE
    >>> Status.UNSTABLE.merge(Status.UNSTABLE)
    UNSTABLE
    >>> Status.UNSTABLE.merge(Status.FAILURE)
    FAILURE

    >>> Status.FAILURE.is_succeeded
    False
    >>> bool(Status.FAILURE)
    False
    >>> Status.FAILURE.merge(Status.SKIPPED)
    FAILURE
    >>> Status.FAILURE.merge(Status.SUCCESS)
    FAILURE
    >>> Status.FAILURE.merge(Status.UNSTABLE)
    FAILURE
    >>> Status.FAILURE.merge(Status.FAILURE)
    FAILURE
    """
    # Numbers stand for the priorities for merging.
    SKIPPED = 0
    SUCCESS = 1
    UNSTABLE = 2
    FAILURE = 3

    @property
    def is_succeeded(self):
        return self.value <= Status.SUCCESS.value

    def merge(self, other: Status):
        return max(self, other, key=lambda status: status.value)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return str(self)

    def __bool__(self) -> bool:
        return self.is_succeeded


def merge_statuses(statuses: Iterable[Status]) -> Status:
    """
    >>> merge_statuses([])
    SKIPPED
    >>> merge_statuses([Status.SUCCESS, Status.UNSTABLE, Status.FAILURE])
    FAILURE
    """
    return reduce(lambda lhs, rhs: lhs.merge(rhs), statuses, Status.SKIPPED)


@dataclass(frozen=True)
class StatusedMixin:
    status: Status = Status.SKIPPED


@dataclass(frozen=True)
class StatusedList(Generic[T], StatusedMixin):
    items: List[T] = field(default_factory=list)


Statused = TypeVar('Statused', bound=StatusedMixin)


def collect_statused(
    items: Iterable[Statused],
) -> StatusedList[Statused]:
    items = list(items)
    status = merge_statuses(item.status for item in items)
    return StatusedList(status=status, items=items)



"""CLI Options."""

import logging
import re
import shlex
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from enum import Enum
from typing import List, Mapping, Optional, Tuple

from ruamel.yaml import safe_load
from ruamel.yaml.error import MarkedYAMLError

from preacher import __version__ as _version


class Level(Enum):
    SKIPPED = logging.DEBUG
    SUCCESS = logging.INFO
    UNSTABLE = logging.WARN
    FAILURE = logging.ERROR

    def __str__(self):
        return self.name.lower()


_LEVEL_MAP: Mapping[str, Level] = {str(level): level for level in Level}


_ENV_PREFIX = 'PREACHER_CLI_'
_ENV_BASE_URL = f'{_ENV_PREFIX}BASE_URL'
_ENV_ARGUMENT = f'{_ENV_PREFIX}ARGUMENT'
_ENV_LEVEL = f'{_ENV_PREFIX}LEVEL'
_ENV_RETRY = f'{_ENV_PREFIX}RETRY'
_ENV_DELAY = f'{_ENV_PREFIX}DELAY'
_ENV_TIMEOUT = f'{_ENV_PREFIX}TIMEOUT'
_ENV_CONCURRENCY = f'{_ENV_PREFIX}CONCURRENCY'
_ENV_REPORT = f'{_ENV_PREFIX}REPORT'

_DEFAULT_ENV_MAP: Mapping[str, Optional[str]] = {
    _ENV_BASE_URL: '',
    _ENV_ARGUMENT: '',
    _ENV_LEVEL: 'success',
    _ENV_RETRY: '0',
    _ENV_DELAY: '0.1',
    _ENV_TIMEOUT: None,
    _ENV_CONCURRENCY: '1',
    _ENV_REPORT: None,
}


def positive_int(value: str) -> int:
    int_value = int(value)
    if int_value <= 0:
        raise ArgumentTypeError(f'must be positive or 0, given {int_value}')
    return int_value


def zero_or_positive_int(value: str) -> int:
    int_value = int(value)
    if int_value < 0:
        raise ArgumentTypeError(f'must be positive or 0, given {int_value}')
    return int_value


def positive_float(value: str) -> float:
    float_value = float(value)
    if float_value <= 0.0:
        raise ArgumentTypeError(f'must be positive, given {float_value}')
    return float_value


def zero_or_positive_float(value: str) -> float:
    float_value = float(value)
    if float_value < 0.0:
        raise ArgumentTypeError(f'must be positive or 0, given {float_value}')
    return float_value


def level(value: str) -> Level:
    result = _LEVEL_MAP.get(value)
    if not result:
        raise ArgumentTypeError(f'invalid level: {value}')
    return result


def argument(value: str) -> Tuple[str, object]:
    match = re.match(r'^([^=]+)=(.*)$', value)
    if not match:
        raise ArgumentTypeError(f'Invalid format argument: {value}')

    key = match.group(1)
    try:
        value = safe_load(match.group(2))
    except MarkedYAMLError as error:
        raise ArgumentTypeError(f'Invalid YAML format. {error}')

    return key, value


def parse_args(
    argv: Optional[List[str]] = None,
    environ: Optional[Mapping[str, str]] = None,
) -> Namespace:
    environ = environ or {}
    defaults = {
        name: environ.get(name) or value
        for (name, value) in _DEFAULT_ENV_MAP.items()
    }

    parser = ArgumentParser()
    parser.add_argument(
        'scenario',
        nargs='*',
        metavar='path',
        help='scenario file paths. When given none, stdin is used instead.'
    )
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=_version,
    )
    parser.add_argument(
        '-u', '--url',
        metavar='url',
        help='specify the base URL',
        default=defaults[_ENV_BASE_URL],
    )
    parser.add_argument(
        '-a', '--argument',
        type=argument,
        metavar='arg',
        nargs='*',
        help='scenario arguments in format "NAME=VALUE" (experimental)',
    )
    parser.add_argument(
        '-l', '--level',
        type=level,
        choices=Level,
        help='show only above or equal to this level',
        default=defaults[_ENV_LEVEL],
    )
    parser.add_argument(
        '-r', '--retry',
        type=zero_or_positive_int,
        metavar='num',
        help='set the max retry count',
        default=defaults[_ENV_RETRY],
    )
    parser.add_argument(
        '-d', '--delay',
        type=zero_or_positive_float,
        metavar='sec',
        help='set the delay between attempts in seconds',
        default=defaults[_ENV_DELAY],
    )
    parser.add_argument(
        '-t', '--timeout',
        type=positive_float,
        metavar='sec',
        help='set the request timeout in seconds',
        default=defaults[_ENV_TIMEOUT],
    )
    parser.add_argument(
        '-c', '--concurrency',
        type=positive_int,
        metavar='num',
        help='set the request concurrency',
        default=defaults[_ENV_CONCURRENCY],
    )
    parser.add_argument(
        '-R', '--report',
        metavar='dir',
        help='report directory',
        default=defaults[_ENV_REPORT],
    )

    args = parser.parse_args(argv)
    args.level = args.level.value

    arguments = args.argument
    if arguments is None:
        try:
            arguments = [
                argument(arg)
                for arg in shlex.split(defaults[_ENV_ARGUMENT] or '')
            ]
        except ValueError as error:
            raise RuntimeError(f'Failed to parse {_ENV_ARGUMENT}: {error}')
    args.argument = dict(arguments)

    return args



from typing import Union

ScalarType = Union[bool, int, float, str]

_SCALAR_TYPES = [bool, int, float, str]


def is_scalar(value: object) -> bool:
    return any(isinstance(value, t) for t in _SCALAR_TYPES)



"""
Test cases, which execute a given request and verify its response
along the given descriptions.
"""

from dataclasses import dataclass, field
from functools import partial
from typing import Optional

from preacher.core.response import Response
from .request import Request
from .response_description import ResponseDescription, ResponseVerification
from .status import Status, StatusedMixin, merge_statuses
from .util.retry import retry_while_false
from .verification import Verification


class CaseListener:
    """
    Interface to listen to running cases.
    Default implementations do nothing.
    """

    def on_response(self, response: Response) -> None:
        pass


@dataclass(frozen=True)
class RequestReport:
    request: Request = field(default_factory=Request)
    result: Verification = field(default_factory=Verification)


@dataclass(frozen=True)
class CaseResult(StatusedMixin):
    """
    Results for the test cases.
    """
    request: RequestReport = field(default_factory=RequestReport)
    response: Optional[ResponseVerification] = None
    label: Optional[str] = None

    def __bool__(self) -> bool:
        return bool(self.status)


class Case:
    """
    Test cases, which execute a given request and verify its response
    along the given descriptions.
    """

    def __init__(
        self,
        label: Optional[str] = None,
        enabled: bool = True,
        request: Optional[Request] = None,
        response: Optional[ResponseDescription] = None,
    ):
        self._label = label
        self._enabled = enabled
        self._request = request or Request()
        self._response = response or ResponseDescription()

    def run(
        self,
        base_url: str = '',
        retry: int = 0,
        delay: float = 0.1,
        timeout: Optional[float] = None,
        listener: Optional[CaseListener] = None,
    ) -> CaseResult:
        if not self._enabled:
            return CaseResult(label=self._label)
        listener = listener or CaseListener()
        func = partial(self._run, base_url, timeout, listener)
        return retry_while_false(func, attempts=retry + 1, delay=delay)

    def _run(
        self,
        base_url: str,
        timeout: Optional[float],
        listener: CaseListener,
    ) -> CaseResult:
        try:
            response = self._request(base_url, timeout=timeout)
        except Exception as error:
            return CaseResult(
                status=Status.FAILURE,
                request=RequestReport(
                    request=self._request,
                    result=Verification.of_error(error),
                ),
                label=self._label,
            )
        listener.on_response(response)

        request_verification = Verification.succeed()
        response_verification = self._response.verify(
            response,
            origin_datetime=response.starts,
        )
        status = merge_statuses([
            request_verification.status,
            response_verification.status,
        ])
        return CaseResult(
            status=status,
            request=RequestReport(
                request=self._request,
                result=request_verification,
            ),
            response=response_verification,
            label=self._label,
        )

    @property
    def label(self) -> Optional[str]:
        return self._label

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def request(self) -> Request:
        return self._request

    @property
    def response(self) -> ResponseDescription:
        return self._response



"""Preacher CLI."""

import logging
import os
import sys
from concurrent.futures import ThreadPoolExecutor
from typing import Iterable

from preacher.app.cli.option import parse_args
from preacher.compilation.factory import create_compiler
from preacher.compilation.yaml import load, load_from_path
from preacher.core.listener.log import LoggingListener
from preacher.core.listener.merging import MergingListener
from preacher.core.listener.report import ReportingListener
from preacher.core.runner import ScenarioRunner
from preacher.presentation.log import ColoredFormatter

FORMATTER = ColoredFormatter()
HANDLER = logging.StreamHandler()
HANDLER.setFormatter(FORMATTER)
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(HANDLER)


def _main() -> None:
    """Main."""
    args = parse_args(environ=os.environ)

    level = args.level
    HANDLER.setLevel(level)
    LOGGER.setLevel(level)

    app = ScenarioRunner(
        base_url=args.url,
        retry=args.retry,
        delay=args.delay,
        timeout=args.timeout,
    )

    if args.scenario:
        objs: Iterable = (load_from_path(path) for path in args.scenario)
    else:
        objs = [load(sys.stdin)]

    compiler = create_compiler()
    scenarios = (
        compiler.compile(obj, arguments=args.argument)
        for obj in objs
    )

    listener = MergingListener()
    listener.append(LoggingListener.from_logger(LOGGER))
    if args.report:
        listener.append(ReportingListener.from_path(args.report))

    with ThreadPoolExecutor(args.concurrency) as executor:
        status = app.run(executor, scenarios, listener=listener)

    if not status.is_succeeded:
        sys.exit(1)


def main():
    try:
        _main()
    except Exception as error:
        LOGGER.exception('%s', error)
        sys.exit(2)
